// SPDX-License-Identifier: GPL-2.0
/*
 * Implementation of the hash table type.
 *
 */
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <net/ipv6.h>
#include "common.h"
#include "hashtab.h"

static struct kmem_cache *hashtab_node_cachep;
struct hashtab *policy_hashtab = NULL;

struct hashtab *hashtab_create(u32 (*hash_value)(struct hashtab *h, const void *key),
			       int (*keycmp)(struct hashtab *h, const void *key1, const void *key2),
			       u32 size)
{
	struct hashtab *p;
	u32 i;

	p = kzalloc(sizeof(*p), GFP_KERNEL);
	if (!p)
		return p;

	p->size = size;
	p->nel = 0;
	p->hash_value = hash_value;
	p->keycmp = keycmp;
	p->htable = kmalloc_array(size, sizeof(*p->htable), GFP_KERNEL);
	if (!p->htable) {
		kfree(p);
		return NULL;
	}

	for (i = 0; i < size; i++)
		p->htable[i] = NULL;
	
	//spin_lock_init(&p->lock);

	return p;
}

u32 hash_value(struct hashtab *h, const void *key) {
    u32 val = 0;
    struct pkt_addr_info *pkt = (struct pkt_addr_info*)key;

    if (pkt->ip_version == IP_ADDR_VERSION_IPV4) {
        val = pkt->saddr.ipv4.s_addr + pkt->daddr.ipv4.s_addr + pkt->sport + pkt->dport + pkt->dir + pkt->ip_version + pkt->protocol;
    } else if (pkt->ip_version == IP_ADDR_VERSION_IPV6) {
        val = ipv6_addr_hash(&pkt->saddr.ipv6) + ipv6_addr_hash(&pkt->daddr.ipv6) + pkt->sport + pkt->dport + pkt->dir + pkt->ip_version + pkt->protocol;
    }
    return val % h->size;
}

int key_cmp(struct hashtab *h, const void *key1, const void *key2)
{	
	struct pkt_addr_info *pkt1 = (struct pkt_addr_info *)key1;
	struct pkt_addr_info *pkt2 = (struct pkt_addr_info *)key2;

	if (!memcmp(&pkt1->sport, &pkt2->sport, sizeof(pkt2->sport)) &&
		!memcmp(&pkt1->dport, &pkt2->dport, sizeof(pkt2->sport)) &&
		pkt1->dir ==  pkt2->dir &&
		pkt1->ip_version == pkt2->ip_version &&
		pkt1->protocol == pkt2->protocol) {
		
		if (pkt1->ip_version == IP_ADDR_VERSION_IPV4) {
			if (!memcmp(&pkt1->saddr.ipv4, &pkt2->saddr.ipv4, sizeof(struct in_addr)) &&
				!memcmp(&pkt1->daddr.ipv4, &pkt2->daddr.ipv4, sizeof(struct in_addr)))
				return 0;
		} else if (pkt1->ip_version == IP_ADDR_VERSION_IPV6) {
			if (!memcmp(&pkt1->saddr.ipv6, &pkt2->saddr.ipv6, sizeof(struct in6_addr)) &&
				!memcmp(&pkt1->daddr.ipv6, &pkt2->daddr.ipv6, sizeof(struct in6_addr))) 
				return 0;
		}
	}		
    /*memcmp比较结构体需要注意字节填充问题，填充的内容可能是随机值*/
	//return memcmp(key1, key2, sizeof(struct pkt_addr_info)) == 0 ? 0 : 1;
	return 1;
}

int hashtab_insert(struct hashtab *h, void *key, unsigned int key_size, 
					void *datum, unsigned int val_size)
{
	u32 hvalue;
	struct hashtab_node *prev, *cur, *newnode;

	//cond_resched();

	if (!h || h->nel == HASHTAB_MAX_NODES)
		return -EINVAL;

	hvalue = h->hash_value(h, key);
	prev = NULL;
	cur = h->htable[hvalue];
    
	while (cur && h->keycmp(h, key, cur->key) == 1) {
		prev = cur;
		cur = cur->next;
	}
	
	if (cur && (h->keycmp(h, key, cur->key) == 0))
		return -EEXIST;
	
	newnode = kmem_cache_zalloc(hashtab_node_cachep, GFP_KERNEL);
	if (!newnode)
		return -ENOMEM;
	
	newnode->key = kmalloc(key_size, GFP_KERNEL);
	newnode->datum = kmalloc(val_size, GFP_KERNEL);
	
	memcpy(newnode->key, key, key_size);
	memcpy(newnode->datum, datum, val_size);

	if (prev) {
		newnode->next = prev->next;
		prev->next = newnode;
	} else {
		newnode->next = h->htable[hvalue];
		h->htable[hvalue] = newnode;
	}

	h->nel++;
	
	return 0;
}


int hashtab_remove(struct hashtab *h, const void *key)
{
	u32 hvalue;
	struct hashtab_node *prev, *cur;

	if (!h)
		return -EINVAL;

	hvalue = h->hash_value(h, key);
	prev = NULL;
	cur = h->htable[hvalue];
	while (cur && h->keycmp(h, key, cur->key) == 1) {
		prev = cur;
		cur = cur->next;
	}

	if (!cur || (h->keycmp(h, key, cur->key) != 0))
		return -ENOENT;

	if (prev)
		prev->next = cur->next;
	else
		h->htable[hvalue] = cur->next;
		
	kfree(cur->key);
	kfree(cur->datum);
	kmem_cache_free(hashtab_node_cachep, cur);

	h->nel--;

	return 0;
}


void *hashtab_search(struct hashtab *h, const void *key)
{
	u32 hvalue;
	struct hashtab_node *cur;

	if (!h)
		return NULL;

	hvalue = h->hash_value(h, key);
	cur = h->htable[hvalue];
	while (cur && h->keycmp(h, key, cur->key) == 1)
		cur = cur->next;

	if (!cur || (h->keycmp(h, key, cur->key) != 0))
		return NULL;

	return cur->datum;
}

void hashtab_cleanup(struct hashtab *h)
{
	u32 i;
	struct hashtab_node *cur, *temp;

	if (!h)
		return;

	for (i = 0; i < h->size; i++) {
		cur = h->htable[i];
		while (cur) {
			temp = cur;
			cur = cur->next;

			kfree(temp->key);
			kfree(temp->datum);
			kmem_cache_free(hashtab_node_cachep, temp);
		}
		h->htable[i] = NULL;
	}

	//kfree(h->htable);
	//h->htable = NULL;

	//kfree(h);
}


void hashtab_destroy(struct hashtab *h)
{
	u32 i;
	struct hashtab_node *cur, *temp;

	if (!h)
		return;

	for (i = 0; i < h->size; i++) {
		cur = h->htable[i];
		while (cur) {
			temp = cur;
			cur = cur->next;
			kfree(temp->key);
			kfree(temp->datum);
			kmem_cache_free(hashtab_node_cachep, temp);
		}
		h->htable[i] = NULL;
	}

	kfree(h->htable);
	h->htable = NULL;

	kfree(h);
}

int hashtab_map(struct hashtab *h,
		int (*apply)(void *k, void *d, void *args),
		void *args)
{
	u32 i;
	int ret;
	struct hashtab_node *cur;

	if (!h)
		return 0;

	for (i = 0; i < h->size; i++) {
		cur = h->htable[i];
		while (cur) {
			ret = apply(cur->key, cur->datum, args);
			if (ret)
				return ret;
			cur = cur->next;
		}
	}
	return 0;
}


void hashtab_stat(struct hashtab *h, struct hashtab_info *info)
{
	u32 i, chain_len, slots_used, max_chain_len;
	struct hashtab_node *cur;

	slots_used = 0;
	max_chain_len = 0;
	for (i = 0; i < h->size; i++) {
		cur = h->htable[i];
		if (cur) {
			slots_used++;
			chain_len = 0;
			while (cur) {
				chain_len++;
				cur = cur->next;
			}

			if (chain_len > max_chain_len)
				max_chain_len = chain_len;
		}
	}

	info->slots_used = slots_used;
	info->max_chain_len = max_chain_len;
}

void __init hashtab_cache_init(void)
{
		hashtab_node_cachep = kmem_cache_create("hashtab_node",
			sizeof(struct hashtab_node),
			0, SLAB_PANIC, NULL);
}

